/*
 * pwm1drv.c
 *   2012/4/12 y.ebihara
 *   2012/5/22 y.ebihara NON-RT Version
 *   2012/6/11 y.ebihara 調光LED用PWMドライバ
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/irq.h>
#include <linux/sh_intc.h>
#include <asm/uaccess.h>
#include <cpu/sh7724.h>

#define DEVNAME "tpu1"

#define BUS_CLOCK	83333333	/* 83MHz */
#define TPU_HZ		100			/* 100Hz */

#define PADR		((volatile unsigned char*)(0xA4050120))		/* ポートA データレジスタ */
#define PACR 		((volatile unsigned short*)(0xA4050100))	/* ポートA 制御レジスタ */
#define PBCR 		((volatile unsigned short*)(0xA4050102))	/* ポートB 制御レジスタ */
#define PSELE		((volatile unsigned short*)(0xA4050156))	/* ピンセレクトレジスタE */
#define MSTPCR2 	((volatile unsigned long*)(0xA4150038))		/* モジュールスタンバイレジスタ２ */
#define TPU_TSTR	((volatile unsigned short*)(0xa4c90000))
#define TPU_BASE0	0xa4c90010
#define TPU_BASE1	0xa4c90050
#define TPU_BASE2	0xa4c90090
#define TPU_BASE3	0xa4c900d0
#define TPU_TCR		0x00
#define TPU_TMDR	0x04
#define TPU_TIOR	0x08
#define TPU_TIER	0x0C
#define TPU_TSR		0x10
#define TPU_TCNT	0x14
#define TPU_TGRA	0x18
#define TPU_TGRB	0x1c
#define TPU_TGRC	0x20
#define TPU_TGRD	0x24

#define IRQ_TPU evt2irq(0x9A0)

struct tpu1drv_dev{
	unsigned long ch;
	unsigned long tpu_base;
	unsigned long interval;
	unsigned long pre_scale;
	unsigned long tgra;
	unsigned long max;
}tpu1drv_dev;

int tpu_hz=TPU_HZ;
module_param(tpu_hz,int, 0644);

/* ============================================================ */
/* ハードウェアレベル                                           */
/* ============================================================ */

static void tpu_outw(struct tpu1drv_dev *dev, int reg, unsigned short value)
{
	volatile unsigned short *addr = (volatile unsigned short*)(dev->tpu_base + reg);
	*addr = value;

	// printk("%s() %08x <= %04x\n", __func__, addr, value);
}

static unsigned short tpu_inw(struct tpu1drv_dev *dev, int reg)
{
	volatile unsigned short *addr = (volatile unsigned short*)(dev->tpu_base + reg);
	return *addr;
}

/* ============================================================ */
/* 割り込みルーチン                                         */
/* ============================================================ */

/* 周期割り込み */
void tpu_tgrb_interrupt(struct tpu1drv_dev *dev)
{
	unsigned short cnt;
	cnt = tpu_inw(dev, TPU_TCNT);

	/* 割り込み遅延最長(ワースト)更新 */
	if(dev->max < cnt){
		dev->max = cnt;
		tpu_outw(dev, TPU_TGRA, cnt); // TPOUT low->high
		printk("%s() max=%d\n",__func__, cnt);
		// WARN_ON(1);	/* レジスタダンプ */
	}

	// currentプロセスを見てみよう
	// printk("SSP=0x%p, thread_info=0x%p, current=0x%p, pid=%d\n",&cnt,current_thread_info(),current,current->pid);
}


static irqreturn_t tpu_handler(int irq, void *_dev)
{
	struct tpu1drv_dev *dev = _dev;
	unsigned short tsr;		/* タイマ ステータス レジスタ */
	tsr = tpu_inw(dev, TPU_TSR);
	if(tsr & (1<<1)){ /* TGRB一致割り込み */
		tsr &= ~(1<<1);
		tpu_outw(dev, TPU_TSR, tsr);  // 要因のクリア
		tpu_tgrb_interrupt(dev);
		return IRQ_HANDLED;	/* 割り込みを処理した */
	}

	printk("%s() error. tsr=0x%x\n",__func__, tsr);
	return IRQ_NONE;	/* 異常 */
}

void tpu1drv_hwinit_channel(struct tpu1drv_dev *dev)
{
	unsigned int tstr_bit = 1 << (dev->ch);

	*TPU_TSTR &= ~tstr_bit;	/* タイマー stop */

	tpu_outw(dev, TPU_TCR,	(2<<5) | dev->pre_scale);		/* TRGBでクリア、 1/1分周 */
	// tpu_outw(dev, TPU_TIOR,	5);					/* High->Low 出力 */
	tpu_outw(dev, TPU_TIOR,	2);					/* Low->High 出力 */
	tpu_outw(dev, TPU_TGRB,	dev->interval);		/* インターバル全体周期 */
	tpu_outw(dev, TPU_TGRA,	dev->tgra);			/* コンペアマッチ   */
	tpu_outw(dev, TPU_TMDR,	2);			/* PWMモード */
	tpu_outw(dev, TPU_TCNT,	0);				/* カウンタ初期化   */
	tpu_outw(dev, TPU_TIER,	(1<<1));				/* TGRB一致割り込み有効 */

	*TPU_TSTR |= tstr_bit;	/* タイマー start */
}

int tpu1drv_hwinit(void)
{
	int interval;
	*MSTPCR2 &= ~(1<<25);	/* モジュールスタンバイレジスタ TPU動作 */
	printk("MSTPCR2=%08x\n",(int)*MSTPCR2);

/* ch1 ピンセレクト */
	*PBCR &= ~(3 << (7*2));	/* PTB7 */

	*PSELE &= ~(3<<14);
	*PSELE |=  (1<<14);		/* TPOUT1,0 選択 */

	tpu1drv_dev.ch = 1;
	tpu1drv_dev.tpu_base = TPU_BASE1;

	interval = BUS_CLOCK/tpu_hz;
	if(interval < 0xffff){
		tpu1drv_dev.interval = interval;
		tpu1drv_dev.pre_scale = 0;
	}else if(interval < 0x3ffff){
		tpu1drv_dev.interval = interval/4;
		tpu1drv_dev.pre_scale = 1;
	}else if(interval < 0xfffff){
		tpu1drv_dev.interval = interval/16;
		tpu1drv_dev.pre_scale = 2;
	}else if(interval < 0x3fffff){
		tpu1drv_dev.interval = interval/64;
		tpu1drv_dev.pre_scale = 3;
	}else{
		printk("%s() tpu_hz=%d\n error\n",__func__,tpu_hz);
		return -EINVAL;
	}

	printk("interval =%ld\n",tpu1drv_dev.interval);
	printk("pre_scale=%ld\n",tpu1drv_dev.pre_scale);

	tpu1drv_dev.tgra = tpu1drv_dev.interval/2;
	tpu1drv_dev.max = 0;

	tpu1drv_hwinit_channel(&tpu1drv_dev);

	return 0;
}

void tpu1drv_hwexit_channel(struct tpu1drv_dev *dev)
{
	unsigned int tstr_bit = 1 << (dev->ch);

	*TPU_TSTR &= ~tstr_bit;	/* タイマー stop */
	tpu_outw(dev, TPU_TIER,	0);	/* 割り込み停止 */
}


void tpu1drv_hwexit(void)
{
	tpu1drv_hwexit_channel(&tpu1drv_dev);

#if 0
	*MSTPCR2 |= (1<<25);	/* モジュールスタンバイレジスタ TPU停止 */
	*TPU_TSTR = 0;	/* タイマー停止 */
#endif
}




/* ============================================================ */
/* file operation                                               */
/* ============================================================ */

int tpu1drv_open(struct inode *inode, struct file *filp)
{
	filp->private_data = (void*)&tpu1drv_dev;
	return 0;
}

static ssize_t tpu1drv_read(struct file *filp, char __user *user_buf,size_t count,loff_t *ppos){
	struct tpu1drv_dev *dev = filp->private_data;
	int ret,len;
	char buffer[128];
	unsigned long scale,usec;

	memset(buffer,0,16);


	if(*ppos>0)
		return 0;

	scale=1;
	if(dev->pre_scale == 1)
		scale = 4;
	else if(dev->pre_scale == 2)
		scale = 16;
	else if(dev->pre_scale == 3)
		scale = 64;

	usec = dev->max;
	usec *= 1000;
	usec /= ((BUS_CLOCK/1000)/scale);

	len=sprintf(buffer,"max=%ld,  %ld,%03ld usec\n",dev->max, (usec/1000),(usec%1000));

	if(len>count)
		len=count;

	ret = copy_to_user(user_buf, buffer, len);
	if(ret<0)
		return ret;

	*ppos += len;
	return len;
}

static ssize_t tpu1drv_write(struct file *filp, const char __user *user_buf,size_t count,loff_t *ppos){
	struct tpu1drv_dev *dev = filp->private_data;
	char str[16];
	int ret,len;
	unsigned long value;

	len=count;
	if(len>15)
		len=15;

	memset(str,0,16);
	ret=copy_from_user(str,user_buf,len);
	if(ret){
		return ret;
	}

	value =  simple_strtoul(str,NULL,0);

	dev->max = value;

	return count;
}

static const struct file_operations tpu1drv_fops = {
        .owner          = THIS_MODULE,
        .write          = tpu1drv_write,
        .read           = tpu1drv_read,
        .open           = tpu1drv_open,
};

static struct miscdevice tpu1drv_miscdev = {
        .minor  = MISC_DYNAMIC_MINOR,	// 動的割り当て
        .name   = DEVNAME,
        .fops   = &tpu1drv_fops,
};


/* ============================================================ */
/* 初期化関数                                                   */
/* ============================================================ */

static int tpu1drv_init(void){
	int ret=0;

	printk("%s() hello. tpu_hz=%d\n",__func__,tpu_hz);

/* 簡単なドライバは miscdev を使うと楽。udev連動してくれる */
	ret = misc_register(&tpu1drv_miscdev);	
	if(ret){
		printk("misc_register() fail. err %d\n",ret);
		goto out1;
	}
	printk("  minor=%d\n",tpu1drv_miscdev.minor);

	/* 割り込み登録 */
	request_irq(IRQ_TPU, tpu_handler, 0, DEVNAME, (void*)&tpu1drv_dev);   /* 割り込みルーチン登録 */

	/* ハードウェア的な初期化 */
	tpu1drv_hwinit();

	return 0;

out1:
	return ret;

}

/* ============================================================ */
/* 終了関数                                                     */
/* ============================================================ */

static void tpu1drv_exit(void){
	printk("%s() goodby\n",__func__);

	/* ハードウェア的な停止 */
	tpu1drv_hwexit();

	/* 登録解除 */
	misc_deregister(&tpu1drv_miscdev);

	/* 割り込み登録の解除 */
	free_irq(IRQ_TPU,(void*)&tpu1drv_dev);

#if 0
/* ポートA */
	*PADR = 0;
	*PACR = 0x5555;	/* PA7-PA0 ポート出力 */
#endif
}


// モジュールの初期化関数、終了関数を定義するマクロ
module_init(tpu1drv_init);
module_exit(tpu1drv_exit);

// モジュールライセンス定義マクロ
MODULE_LICENSE("GPL");

