/*
 * tpudriver_wakeup.c
 *   2013/
 *   2013/10/22 y.ebihara wakeup
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/irq.h>
#include <linux/sh_intc.h>
#include <linux/interrupt.h>   // for tasklets
#include <linux/workqueue.h>   // for workqueue
#include <asm/uaccess.h>
#include <asm/io.h>
#include <cpu/sh7724.h>

/* -----------------------------------
 * 低レベル IOポート入出力関数
 * ----------------------------------- */

#define PMOD    (0)     /* その他の機能 */
#define POUT    (1) /* 出力 */
#define PINP    (2) /* 入力 pullup 有り */
#define PIN_    (3) /* 入力 pullup 無し */

#define PADR 0xA4050120
#define PBDR 0xA4050122
#define PGDR 0xA405012C
#define PJDR 0xA4050130
#define PRDR 0xA405013C
#define PUDR 0xA4050162
#define PACR (PADR-0x20)
#define HIZCRA 0xA4050158

#define DEVNAME "tpudrv"

#define BUS_CLOCK	83333333	/* 83MHz */
#define TPU_HZ		25			/* 25Hz */

#define PSELE		((volatile unsigned short*)(0xA4050156))	/* ピンセレクトレジスタE */
#define MSTPCR2 	((volatile unsigned long*)(0xA4150038))		/* モジュールスタンバイレジスタ２ */
#define TPU_TSTR	((volatile unsigned short*)(0xa4c90000))
#define TPU_BASE0	0xa4c90010
#define TPU_BASE1	0xa4c90050
#define TPU_BASE2	0xa4c90090
#define TPU_BASE3	0xa4c900d0
#define TPU_TCR		0x00
#define TPU_TMDR	0x04
#define TPU_TIOR	0x08
#define TPU_TIER	0x0C
#define TPU_TSR		0x10
#define TPU_TCNT	0x14
#define TPU_TGRA	0x18
#define TPU_TGRB	0x1c
#define TPU_TGRC	0x20
#define TPU_TGRD	0x24

#define IRQ_TPU evt2irq(0x9A0)

struct tpudrv_dev{
	unsigned long ch;
	unsigned long tpu_base;
	unsigned long interval;
	unsigned long pre_scale;
	unsigned long tgra;
	unsigned long max;
}tpudrv_dev[4];

static wait_queue_head_t tpudrv_wq;
static unsigned long cnt0_save;		/* 割込み発生タイミングを控える */
static unsigned long delay_max =0;

int tpu_hz=TPU_HZ;
module_param(tpu_hz,int, 0644);


/* ============================================================ */
/* ハードウェアレベル                                           */
/* ============================================================ */

static void port_control_bit(unsigned long address, int bit, int mod){
	unsigned short tmp,mask;
	mask = 3 << (bit*2);
	tmp = *((volatile unsigned short*)(address));
	tmp &= ~mask;
	tmp |= (mod)<<(bit*2);
	*((volatile unsigned short*)(address)) = tmp;
};

static void tpu_outw(struct tpudrv_dev *dev, int reg, unsigned short value)
{
	volatile unsigned short *addr = (volatile unsigned short*)(dev->tpu_base + reg);
	*addr = value;

	// printk("%s() %08x <= %04x\n", __func__, addr, value);
}

static unsigned short tpu_inw(struct tpudrv_dev *dev, int reg)
{
	volatile unsigned short *addr = (volatile unsigned short*)(dev->tpu_base + reg);
	return *addr;
}

static void gpio_high(unsigned long address, int bit)
{
	*((volatile unsigned char*)address) |= (1<<bit);
}

static void gpio_low(unsigned long address, int bit)
{
	*((volatile unsigned char*)address) &= ~(1<<bit);
}


/* ============================================================ */
/* 割り込みルーチン                                             */
/* ============================================================ */

/* TPU0 周期割り込み */
void tpu0_tgrb_isr(struct tpudrv_dev *dev)
{
	unsigned short cnt;
	cnt = tpu_inw(dev, TPU_TCNT);
	cnt0_save=cnt;	/* 割込み発生タイミングを控える */

	/* 割り込み遅延最長(ワースト)更新 */
	if(dev->max < cnt){
		dev->max = cnt;
		tpu_outw(dev, TPU_TGRA, cnt); // TPOUT low->high
		printk("%s() max=%d\n",__func__, cnt);
		// WARN_ON(1);	/* レジスタダンプ */
	}

	wake_up_interruptible(&tpudrv_wq);	/* read()で寝ているプロセスを起床 */
}

/* 割込みサービスルーチン Top Half */
static irqreturn_t tpu_handler(int irq, void *_dontuse)
{
	struct tpudrv_dev *dev;
	unsigned short tsr;		/* タイマ ステータス レジスタ */

	dev = &tpudrv_dev[0];	/* TPU0 */
	tsr = tpu_inw(dev, TPU_TSR);
	if(tsr & (1<<1)){ /* TGRB一致割り込み */
		tsr &= ~(1<<1);
		tpu_outw(dev, TPU_TSR, tsr);  // 要因のクリア
		tpu0_tgrb_isr(dev);
		return IRQ_HANDLED;
	}

	printk("%s() error. tsr=0x%x\n",__func__, tsr);
	return IRQ_NONE;	/* 異常 */
}

/* ============================================================ */
/* ハードウェア初期化                                           */
/* ============================================================ */

int tpudrv_hwinit_channel(struct tpudrv_dev *dev, int tpu_hz, int interrupt_enable)
{
	int interval;
	unsigned int tstr_bit = 1 << (dev->ch);

	interval = BUS_CLOCK/tpu_hz;
	if(interval < 0xffff){
		dev->interval = interval;
		dev->pre_scale = 0;
	}else if(interval < 0x3ffff){
		dev->interval = interval/4;
		dev->pre_scale = 1;
	}else if(interval < 0xfffff){
		dev->interval = interval/16;
		dev->pre_scale = 2;
	}else if(interval < 0x3fffff){
		dev->interval = interval/64;
		dev->pre_scale = 3;
	}else{
		printk("%s() tpu_hz=%d\n error\n",__func__,tpu_hz);
		return -EINVAL;
	}

	printk("interval =%ld\n",dev->interval);
	printk("pre_scale=%ld\n",dev->pre_scale);

	dev->tgra = dev->interval/2;
	dev->max = 0;

	*TPU_TSTR &= ~tstr_bit;	/* タイマー stop */

	tpu_outw(dev, TPU_TCR,	(2<<5) | dev->pre_scale);		/* TRGBでクリア、 1/1分周 */
	// tpu_outw(dev, TPU_TIOR,	5);					/* High->Low 出力 */
	tpu_outw(dev, TPU_TIOR,	2);					/* Low->High 出力 */
	tpu_outw(dev, TPU_TGRB,	dev->interval);		/* インターバル全体周期 */
	tpu_outw(dev, TPU_TGRA,	dev->tgra);			/* コンペアマッチ   */
	tpu_outw(dev, TPU_TMDR,	2);			/* PWMモード */
	tpu_outw(dev, TPU_TCNT,	0);				/* カウンタ初期化   */
	tpu_outw(dev, TPU_TIER,	interrupt_enable);				/* 割込み有効 */

	// *TPU_TSTR |= tstr_bit;	/* タイマー start */

	return 0;
}

int tpudrv_hwinit(void)
{
	*MSTPCR2 &= ~(1<<25);	/* モジュールスタンバイレジスタ TPU動作 */
	printk("MSTPCR2=%08x\n",(int)*MSTPCR2);

	port_control_bit(PBDR-0x20, 6, PMOD);	// PTB6 .. module
	port_control_bit(PBDR-0x20, 7, PMOD);	// PTB7 .. module

	*PSELE &= ~(3<<14);
	*PSELE |=  (1<<14);		/* TPOUT1,0 選択 */

	tpudrv_dev[0].ch=0;
	tpudrv_dev[0].tpu_base=TPU_BASE0;

	tpudrv_dev[1].ch=1;
	tpudrv_dev[1].tpu_base=TPU_BASE1;

	tpudrv_dev[2].ch=2;
	tpudrv_dev[2].tpu_base=TPU_BASE2;

	tpudrv_dev[3].ch=3;
	tpudrv_dev[3].tpu_base=TPU_BASE3;

	*TPU_TSTR = 0;		// TPU0, TPU1タイマー停止

	tpudrv_hwinit_channel(&tpudrv_dev[0], tpu_hz, (1<<1));	/* TPU0 割込み有効 */
	tpudrv_hwinit_channel(&tpudrv_dev[1], tpu_hz, 0);		/* TPU1 割込み無効 */

	*TPU_TSTR = (1<<0)|(1<<1);		// TPU0, TPU1タイマースタート

	return 0;
}

void tpudrv_hwexit_channel(struct tpudrv_dev *dev)
{
	unsigned int tstr_bit = 1 << (dev->ch);

	*TPU_TSTR &= ~tstr_bit;	/* タイマー stop */
	tpu_outw(dev, TPU_TIER,	0);	/* 割り込み停止 */
}


void tpudrv_hwexit(void)
{
	tpudrv_hwexit_channel(&tpudrv_dev[0]);
	tpudrv_hwexit_channel(&tpudrv_dev[1]);

#if 0
	*MSTPCR2 |= (1<<25);	/* モジュールスタンバイレジスタ TPU停止 */
	*TPU_TSTR = 0;	/* タイマー停止 */
#endif
}


/* ============================================================ */
/* file operation                                               */
/* ============================================================ */

int tpudrv_open(struct inode *inode, struct file *filp)
{
	return 0;
}

int tpudrv_calc_usec(struct tpudrv_dev *dev)
{
	unsigned long scale,usec;

	scale=1;
	if(dev->pre_scale == 1)
		scale = 4;
	else if(dev->pre_scale == 2)
		scale = 16;
	else if(dev->pre_scale == 3)
		scale = 64;

	usec = dev->max;
	usec *= 1000;
	usec /= ((BUS_CLOCK/1000)/scale);

	return usec;
}


static ssize_t tpudrv_read(struct file *filp, char __user *user_buf,size_t count,loff_t *ppos){
	int ret,len;
	unsigned long usec0, usec1, cnt, delay;
	struct tpudrv_dev *dev0, *dev1;
	char buffer[128];

	memset(buffer,0,sizeof(buffer));

#if 0
	if(*ppos>0)
		return 0;
#endif

	interruptible_sleep_on(&tpudrv_wq);	/* 割込みが入るまでスリープする */
	if(signal_pending(current)){
		return -ERESTARTSYS;	/* 異常起床 (CTRL+C が押されるなどで Signalが発生した場合) */
	}

	dev0 = &tpudrv_dev[0];	/* TPU0 */
	dev1 = &tpudrv_dev[1];	/* TPU1 */

	cnt = tpu_inw(dev0, TPU_TCNT);	/* TPU0 カウンタ値を得る */
	delay = cnt-cnt0_save;

	/* 最大更新なら控える */
	if(delay_max < delay){
		delay_max = delay;
		tpu_outw(dev1, TPU_TGRA, cnt); // TPOUT low->high
		printk("%s() max=%ld\n",__func__, delay);
		dev1->max = delay;
	}

	usec0 = tpudrv_calc_usec(&tpudrv_dev[0]);
	usec1 = tpudrv_calc_usec(&tpudrv_dev[1]);

	len=sprintf(buffer,"ISR delay=%ld,%03ld usec,  wakeup delay=%ld,%03ld usec\n",
		(usec0/1000),(usec0%1000),
		(usec1/1000),(usec1%1000));

	if(len>count)
		len=count;

	ret = copy_to_user(user_buf, buffer, len);
	if(ret<0)
		return ret;

	*ppos += len;
	return len;
}

static const struct file_operations tpudrv_fops = {
        .owner          = THIS_MODULE,
        .read           = tpudrv_read,
        .open           = tpudrv_open,
};

static struct miscdevice tpudrv_miscdev = {
        .minor  = MISC_DYNAMIC_MINOR,	// 動的割り当て
        .name   = DEVNAME,
        .fops   = &tpudrv_fops,
};


/* ============================================================ */
/* 初期化関数                                                   */
/* ============================================================ */

static int tpudrv_init(void){
	int ret=0;

	printk("%s() hello. tpu_hz=%d\n",__func__,tpu_hz);

	/* 割り込み登録 */
	request_irq(IRQ_TPU, tpu_handler, 0, DEVNAME, (void*)tpudrv_dev);   /* 割り込みルーチン登録 */

	/* 待ち列の初期化 */
	init_waitqueue_head(&tpudrv_wq);

	/* ハードウェア的な初期化 */
	tpudrv_hwinit();

/* 簡単なドライバは miscdev を使うと楽。udev連動してくれる */
	ret = misc_register(&tpudrv_miscdev);	
	if(ret){
		printk("misc_register() fail. err %d\n",ret);
		goto out1;
	}
	printk("  minor=%d\n",tpudrv_miscdev.minor);

	return 0;

out1:
	return ret;

}

/* ============================================================ */
/* 終了関数                                                     */
/* ============================================================ */

static void tpudrv_exit(void){
	printk("%s() goodby\n",__func__);

	/* ハードウェア的な停止 */
	tpudrv_hwexit();

	/* 登録解除 */
	misc_deregister(&tpudrv_miscdev);

	/* 割り込み登録の解除 */
	free_irq(IRQ_TPU,(void*)&tpudrv_dev);

/* ポートA */
	writeb(0xFF, PADR);
	writew(0x5555, PACR);
}


// モジュールの初期化関数、終了関数を定義するマクロ
module_init(tpudrv_init);
module_exit(tpudrv_exit);

// モジュールライセンス定義マクロ
MODULE_LICENSE("GPL");

