/*
 * malloc & free
 *
 */

#include "malloc.h"


/// 管理ヘッダ. sizeof(Header) == MALLOC_UNIT_SZ を満たすこと
typedef union Header {
	struct {
		union Header *ptr;          // 次のブロックへのポインタ
		unsigned int size;          // サイズ(sizeof(Header)単位の個数)
	} s;
	char x[MALLOC_UNIT_SZ];			 // アライメント/基本サイズの調整
} Header;

/// 空きブロックの管理用
static Header *base = 0;
static Header *freep = 0;

/// アドレスmemから sizeバイトをmallocで使うヒープ・メモリとして初期化.
void malloc_init(void *mem, unsigned int size)
{
    // 先頭アドレスをMALLOC_UNIT_SZ単位にアライメントしておく
	base          = (Header*)(((unsigned long)mem+MALLOC_UNIT_SZ-1) & ~(MALLOC_UNIT_SZ-1));
	base->s.ptr   = base + 1;
	base->s.size  = 0;
	freep         = base->s.ptr;
	freep->s.ptr  = base;
	// アドレス調整を考慮してサイズを求める. サイズは'Header'個数。
	freep->s.size = ((char*)mem + size - (char*)freep) / MALLOC_UNIT_SZ;
}

/// nbytesのメモリをヒープから確保
void *malloc(unsigned int nbytes)
{
    // ヘッダ部を含めた必要メモリのサイズ(Header単位)を求める
    unsigned nunits = (nbytes + MALLOC_UNIT_SZ-1) / MALLOC_UNIT_SZ + 1;
    Header   *prevp = freep;
    Header   *p;

    // 空きブロック一覧を順繰りに調べる
    for (p = prevp->s.ptr; ;prevp = p, p = p->s.ptr) {
        if (p->s.size >= nunits)    // 要求サイズを満たすブロックがあった
            break;
        if (p == freep)             // 空き一覧の先頭に戻ったのなら
            return (void*)0;            // 空きが無かった
    }
    // 空きが見つかったときの処理
    if (p->s.size == nunits) {      // 見つかったブロックが調度のとき
        prevp->s.ptr = p->s.ptr;    // 空き一覧リンクはら外す
    } else {
        p->s.size -= nunits;
        p         += p->s.size;
        p->s.size  = nunits;
    }
    freep = prevp;                  // 今回の結果を空き一覧の先頭に反映.
    return (void *)(p + 1);         // ヘッダ部を除いたアドレスを返す
}

/// malloc したメモリを開放
void free(void *ap)
{
    Header *p;
    Header *bp = (Header *)ap - 1;

    // s.ptr は低いアドレスから高アドレスにむかって繋がっている
    // 返却されたメモリが、空きブロックリストのどのへんに入るか、探す
    for (p = freep; !(p < bp && bp < p->s.ptr); p = p->s.ptr) {
        if (p >= p->s.ptr && (bp > p || bp < p->s.ptr))
            break;
    }
    // 以下、挿入場所が見つかったものとして処理
    if (bp + bp->s.size == p->s.ptr) {  // 直後の要素と合体できる場合
        bp->s.size += p->s.ptr->s.size; //   サイズ更新
        bp->s.ptr  =  p->s.ptr->s.ptr;  //   繋ぎ替え
    } else {                            // 合体できない場合
        bp->s.ptr  =  p->s.ptr;         //   繋ぎ替え
    }
    if (p + p->s.size == bp) {          // 直前の要素と合体できる場合
        p->s.size += bp->s.size;        //   サイズ更新
        p->s.ptr  =  bp->s.ptr;         //   繋ぎ替え
    } else {                            // 合体できない場合
        p->s.ptr  =  bp;                //   繋ぎ替え
    }
    freep = p;                          // 空きブロックの先頭を更新
}

