/*
 * dmadrv.c
 *   for linux-3.x kernel
 *   written by Y.Ebihara
 *
 *  4Mバイトのデータを PortA DataRegisterにDMA転送（出力）する
 *
 *  ドライバロード時に 4Mバイトを確保
 *  userプログラムから mmap() でデータを書き込み
 *  userプログラムから ioctl() でDMA転送開始指示
 *
 *   2014/8/30  Version 1.0
 */


#include <linux/module.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/dma-mapping.h>
#include <asm/uaccess.h>

#define DMADRV_MAJOR 249
#define DEVNAME "dmadrv"

#define BIGSIZE_ORDER 10	/* 4Mbyte */

struct dmadrv_dev{
	unsigned long base;			/* ベースアドレス */
	unsigned long size;			/* バッファの大きさ */
	unsigned long data_len;		/* データの大きさ */
	struct mutex  mutex;		/* ロック */
}dmadrv_dev;

#define DMADRV_DMA_START 0x1000   /* Start DMA  */
#define DMADRV_DMA_STOP  0x1001   /* Stop  DMA  */

/* ======================================================== */
/* DMA_transmit                                             */
/* ======================================================== */

#define PACR   0xA4050100 /* ポートA コントロールレジスタ */
#define PADR   0xA4050120 /* ポートA データレジスタ */

#define MSTPCR0  0xA4150030	/* モジュールストップレジスタ0 */

#define DMA0_DMAOR  0xFE008060 // DMA0 オペレーションレジスタ                16

#define DMA0_SAR_0  0xFE008020 // DMA0 ソースアドレスレジスタ_0              32
#define DMA0_DAR_0  0xFE008024 // DMA0 デスティネーションアドレスレジスタ_0  32
#define DMA0_TCR_0  0xFE008028 // DMA0 トランスファカウントレジスタ_0        32
#define DMA0_SARB_0 0xFE008120 // DMA0 ソースアドレスレジスタB_0             32
#define DMA0_DARB_0 0xFE008124 // DMA0 デスティネーションアドレスレジスタB_0 32
#define DMA0_TCRB_0 0xFE008128 // DMA0 トランスファカウントレジスタB_0       32

#define DMA0_CHCR_0 0xFE00802C // DMA0 チャネルコントロールレジスタ_0        32


#define DMA_DMAOR  ((volatile unsigned short*)DMA0_DMAOR)
#define DMA_SAR    ((volatile unsigned long*)DMA0_SAR_0)
#define DMA_DAR    ((volatile unsigned long*)DMA0_DAR_0)
#define DMA_TCR    ((volatile unsigned long*)DMA0_TCR_0)
#define DMA_SARB   ((volatile unsigned long*)DMA0_SARB_0)
#define DMA_DARB   ((volatile unsigned long*)DMA0_DARB_0)
#define DMA_TCRB   ((volatile unsigned long*)DMA0_TCRB_0)
#define DMA_CHCR   ((volatile unsigned long*)DMA0_CHCR_0)

void dmadrv_hwinit(void)
{
	*(volatile unsigned short*)PACR = 0x0055;	/* 0b 00000000 01010101 */

	/* 省電力レジスタ/クロック供給ON */
	*(volatile unsigned long*)MSTPCR0 &= ~(1<<21);	/* DMAC0 */
	*(volatile unsigned long*)MSTPCR0 &= ~(1<<12);  /* DMAC1 */

	printk("*MSTPCR0 = 0x%lx\n",*(volatile unsigned long*)MSTPCR0);
}

int dmadrv_sh4a_dma_tx_init(struct dmadrv_dev *dev)
{
	unsigned long src, dst,len; /* 転送元、転送先 */

	src = dev->base;	/* 転送元: 確保したメモリ */
	dst = PADR;	        /* 転送先: PortA データレジスタ */
	len = dev->data_len; /* 転送サイズ: データ長 */

	*DMA_SAR = src;	/* 転送元 */
	*DMA_DAR = dst;	/* 転送先 */
	*DMA_TCR = len; /* 長さ */

#define CHCR_LCKN  0
#define CHCR_RPT   0x1  /* リピートモード使用 */
#define CHCR_TS    0    /* バイト単位 */
#define CHCR_DM    0    /* 転送先固定 */
#define CHCR_SM    1    /* 転送元インクリメント */
#define CHCR_RS    4    /* オートリクエスト */
#define CHCR_TB    0    /* サイクルスチールモード */
#define CHCR_IE    0    /* 割り込み */

	*DMA_CHCR = (CHCR_LCKN<<30) | (CHCR_RPT<<25) | (((CHCR_TS>>2)&0x03)<<20) | 
	            (CHCR_DM<<14) | (CHCR_SM<<12) | (CHCR_RS<<8) | (CHCR_TB<<5) |
	            ( (CHCR_TS&0x03) <<3) | (CHCR_IE<<2);

#define DMAOR_CMS 0x04	/* インタミテットモード256 */

	*DMA_DMAOR = (DMAOR_CMS<<12) | 1 ;

	return 0; /* success */
}

void dmadrv_sh4a_debugprint(void)
{
	printk("DMA_DMAOR 0x%x\n", *DMA_DMAOR );
	printk("DMA_SAR   0x%lx\n",*DMA_SAR   );
	printk("DMA_DAR   0x%lx\n",*DMA_DAR   );
	printk("DMA_TCR   0x%lx\n",*DMA_TCR   );
	printk("DMA_SARB  0x%lx\n",*DMA_SARB  );
	printk("DMA_DARB  0x%lx\n",*DMA_DARB  );
	printk("DMA_TCRB  0x%lx\n",*DMA_TCRB  );
	printk("DMA_CHCR  0x%lx\n",*DMA_CHCR  );
}

int dmadrv_sh4a_dma_start(struct dmadrv_dev *dev)
{
	printk("%s()\n",__func__);

	dmadrv_sh4a_dma_tx_init(dev);

	/* DMA転送メモリのキャッシュを書き出す */
	dma_cache_sync(NULL, (void*)dev->base, dev->size, DMA_TO_DEVICE);

	*DMA_CHCR &= ~0x02;		/* clear TE bit */
	*DMA_CHCR |= 0x01;

	dmadrv_sh4a_debugprint();
	return 0; /* success */
}

int dmadrv_sh4a_dma_stop(struct dmadrv_dev *dev)
{
	printk("%s()\n",__func__);
	*DMA_CHCR &= ~0x01;	/* clear DE bit */

	// dmadrv_sh4a_debugprint();
	return 0; /* success */
}


/* ======================================================== */
/* file operations                                          */
/* ======================================================== */

static int dmadrv_open(struct inode *inode, struct file *filp)
{
	filp->private_data = &dmadrv_dev;
	return 0;
}

static int dmadrv_read(struct file *filp, char *user_buf, size_t count, loff_t *ppos)
{
	int ret=0;
	struct dmadrv_dev *dev = filp->private_data;

	if(count > dev->data_len - *ppos)
		count = dev->data_len - *ppos;

	mutex_lock(&dev->mutex);

	ret = copy_to_user(user_buf, (void*)(dev->base) + *ppos, count);
	if(ret){
		goto out1;
	}

	*ppos += count;
	ret = count;

out1:
	mutex_unlock(&dev->mutex);
	return ret;
}

static int dmadrv_write(struct file *filp, const char *user_buf,size_t count,loff_t *ppos)
{
	int ret=0;
	struct dmadrv_dev *dev = filp->private_data;

	if(count > dev->size - *ppos){
		count = dev->size - *ppos;
	}

	mutex_lock(&dev->mutex);

	ret=copy_from_user((void*)(dev->base) + *ppos, user_buf, count);
	if(ret){
		goto out1;
	}

	*ppos += count;
	dev->data_len = *ppos;
	ret = count;

out1:
	mutex_unlock(&dev->mutex);
	return ret;
}

static int dmadrv_mmap(struct file *filp, struct vm_area_struct *vma)
{
	int ret=0;
	struct dmadrv_dev *dev = filp->private_data;

	if ((vma->vm_end - vma->vm_start) > dev->size) {
		printk("vma size (0x%lx) is too big.\n",(vma->vm_end - vma->vm_start));
		return -EINVAL;
	}

	ret = remap_pfn_range(
			vma,						// 仮想メモリ(VMA)構造体
			vma->vm_start,				// 仮想メモリの開始番地
			dev->base >> PAGE_SHIFT,	// 物理メモリのページ番号
			vma->vm_end - vma->vm_start,	// メモリサイズ
			vma->vm_page_prot);				// ページアトリビューション
	if(ret){
		printk("remap_page_range() failed. %d\n",ret);
		return ret;
	}

	return 0;
}

static long dmadrv_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret=0;
	struct dmadrv_dev *dev = filp->private_data;

	printk("%s()\n",__func__);

	switch(cmd){
		case DMADRV_DMA_START:
			dev->data_len = arg;
			dmadrv_sh4a_dma_start(dev);
			break;
		case DMADRV_DMA_STOP:
			dmadrv_sh4a_dma_stop(dev);
			break;
		default:
			printk("%s() cmd=%d not implement\n",__func__,cmd);
			ret=-EINVAL;
	}

	return ret;
}


static const struct file_operations dmadrv_fops = {
	.owner          = THIS_MODULE,
	.open	= dmadrv_open,
	.read	= dmadrv_read,
	.write	= dmadrv_write,
	.mmap	= dmadrv_mmap,
	.unlocked_ioctl = dmadrv_ioctl,
};

static struct miscdevice dmadrv_miscdev = {
	.minor  = MISC_DYNAMIC_MINOR,
	.name   = DEVNAME,
	.fops   = &dmadrv_fops,
};

static int dmadrv_init(void){
	int ret=0;
	struct dmadrv_dev *dev;
	printk("%s() hello\n",__func__);

	dmadrv_hwinit();

	dev = &dmadrv_dev;

	/* 物理メモリを直接確保する */
	dev->base = __get_free_pages(GFP_KERNEL | GFP_DMA, BIGSIZE_ORDER);
	dev->size = PAGE_SIZE*((2<<BIGSIZE_ORDER)/2);
	if(dev->base == 0){
		printk("%s() -ENOMEM\n",__func__);
		return -ENOMEM;	/* out of memory */
	}
	printk("kernel_buffer=%lx size=%ld\n",dev->base, dev->size);

	/* キャラクタ型ドライバの登録（簡易） */
	ret = misc_register(&dmadrv_miscdev);
	if(ret){
		printk("misc_register() fail. err %d\n",ret);
		goto out1;
	}
	printk("  minor=%d\n",dmadrv_miscdev.minor);

	/* ロックの初期化 */
	mutex_init(&dev->mutex);

	return 0;

out1:
	free_pages(dev->base, BIGSIZE_ORDER);

	return ret;
}

static void dmadrv_exit(void){
	struct dmadrv_dev *dev = &dmadrv_dev;
	printk("%s() goodby\n",__func__);

	if(dev->base)
		free_pages(dev->base,BIGSIZE_ORDER);

	misc_deregister(&dmadrv_miscdev);
}

// モジュールの初期化関数、終了関数を定義するマクロ
module_init(dmadrv_init);
module_exit(dmadrv_exit);

// モジュールライセンス定義マクロ
MODULE_LICENSE("GPL");
